'use strict';

var algorithms = {

  where: function(query, table, where, count) {
    count = (typeof count === 'number') ? count : 0;
    let operator = '=';
    for(let column in where) {
      if(!count) {
        query = query.where([table, column], operator, where[column]);
      } else {
        query = query.and([table, column], operator, where[column]);
      }
      count++;
    }
    return {
      query: query,
      count: count
    };
  }

};

var Query = {

  /**
   * Create
   * @param {Object} params
   * @param {Object} params.ORM
   * @param {string} params.table
   * @param {Object} params.data
   * @param {Object} params.returning
   */
  create: function(params) {
    return params.ORM.querier.table(params.table)
    .insert([params.data])
    .returning(params.returning)
    .exec();
  },

  /**
   * Update
   * @param {Object} params
   * @param {Object} params.ORM
   * @param {string} params.table
   * @param {Object} params.where
   * @param {Object} params.data
   * @param {Object} params.returning
   */
  update: function(params) {
    let query = params.ORM.querier.table(params.table);
    query = query.update(params.data);
    let whereCount = 0;
    let results = algorithms.where(query, params.table, params.where, whereCount);
    whereCount += results.count;
    query = results.query;
    query = query.returning(params.returning);
    return query.exec();
  },

  /**
   * Find
   * @param {Object} params
   * @param {Object} params.ORM
   * @param {string} params.table
   * @param {Object} params.join
   * @param {Object} params.where
   * @param {Object} params.relationships
   * @param {Object} params.options
   * @param {Object} params.returning
   */
  find: function(params) {
    return new Promise((resolve, reject) => {
      //
      // Query
      let query = params.ORM.querier.table(params.table)
      .distinct()
      .select([[params.table, '*']]);
      //
      // Join
      for(let table in params.join) {
        let relationship = params.relationships[table];
        let throughTable = relationship.getThroughTable();
        let parent = (relationship.type === 'BELONGS_TO_MANY')
        ? relationship.identifer.parent_0
        : relationship.identifer.parent;
        let child = (relationship.type === 'BELONGS_TO_MANY')
        ? relationship.identifer.child_0
        : relationship.identifer.child;
        query.innerJoin(throughTable).on([throughTable, parent], '=', [params.table, child]);
      }
      // Where
      let whereCount = 0;
      // Where table attributes
      let results = algorithms.where(query, params.table, params.where, whereCount);
      whereCount += results.count;
      query = results.query;
      // Where join attributes
      for(let table in params.join) {
        let relationship = params.relationships[table];
        let throughTable = relationship.getThroughTable();
        //
        // BelongsToMany
        if(relationship.type === 'BELONGS_TO_MANY') {
          let relationTable = relationship.modelFactoryName;
          let parent_0 = relationship.identifer.parent_0;
          let child_0 = relationship.identifer.child_0;
          let parent_1 = relationship.identifer.parent_1;
          let child_1 = relationship.identifer.child_1;
          // Relation select
          let relation = params.ORM.querier.table(throughTable)
          .select([[throughTable, parent_0]])
          .innerJoin(relationTable)
          .on([throughTable, parent_1], '=', [relationTable, child_1]);
          // Relation where
          let joinWhereCount = 0;
          params.join[table].map((table) => {
            // Where table attributes
            let results = algorithms.where(relation, relationTable, table, joinWhereCount);
            joinWhereCount += results.count;
            relation = results.query;
          });
          // In
          if(!whereCount) {
            query = query.where([params.table, child_0]);
          } else {
            query = query.and([params.table, child_0]);
          }
          query = query.in(relation.sql());
          whereCount++;
        }
        //
        // BelongsToOne
        // hasOne
        // hasMany
        else {
          // Where table attributes
          let results = algorithms.where(query, table, params.join[table], whereCount);
          whereCount += results.count;
          query = results.query;
        }
      }
      //
      // Execute
      query.exec()
      .then((results) => resolve(results))
      .catch((error) => reject(error));
    });
  },

  /**
   * Delete
   * @param {Object} params
   * @param {Object} params.ORM
   * @param {string} params.table
   * @param {Object} params.where
   * @param {Object} params.returning
   */
  delete: function(params) {
    let query = params.ORM.querier.table(params.table);
    query = query.delete();
    let whereCount = 0;
    let results = algorithms.where(query, params.table, params.where, whereCount);
    whereCount += results.count;
    query = results.query;
    query = query.returning(params.returning);
    return query.exec();
  }

};

module.exports = Query;